<?php

/**
 * The widget-related functionality of the plugin.
 * 
 * Defines the plugin name, version, and methods that adds the widget to the 
 * WordPress administration dashboard.
 *
 * @link       https://cobweb-security.com
 * @since      4.1.6
 *
 * @package    Cwis
 * @subpackage Cwis/includes/application
 * @author     Cobweb Security <cwis@cobweb-security.com>
 */
class Cwis_Application_Widget extends Cwis_Core_Foundation
{

    /**
     * The URI of the RSS feed we want to retrieve. If an array of URIs,
     * the feeds are merged using SimplePie's multifeed feature.
     * 
     * @var string|array
     */
    protected $feed_uri = array(
        'https://cobweb-security.com/laboratory/feed',
    );

    public function __construct($plugin_name, $version, $loader = null)
    {
        parent::__construct($plugin_name, $version, $loader);
    }

    /**
     * Adds widget to the WP-admin dashboard.
     *
     * @since   4.1.6
     * @access  public
     * 
     * @return  string|void
     */
    public function add_widget()
    {
        global $wp_meta_boxes;
        if (isset($wp_meta_boxes['dashboard']['normal']['core']['cobwebsec'])) {
            return;
        }

        wp_add_dashboard_widget('cobwebsec', 'WebDefender Security (formely CWIS Antivirus)',
                                array(&$this, 'render_dashboard_widget'));
    }

    /**
     * Renders widget content.
     *
     * @since   4.1.6
     * @access  public
     */
    public function render_dashboard_widget()
    {
        // Get parsed and cached RSS feed items
        $rss_items = $this->get_rss_items();

        // Generate links to plugin page
        $cwis_gdpr_url = 'admin.php?page=cwis-gdpr';
        $cwis_scanner_url = 'admin.php?page=cwis-scanner';
        $cwis_updater_url = 'admin.php?page=cwis-updater';
        if (function_exists('network_admin_url')) {
            $cwis_scanner_url = network_admin_url($cwis_scanner_url);
            $cwis_updater_url = network_admin_url($cwis_updater_url);
        }

        // Get plugin module statuses
        $cwis_defender_settings = $this->get_class_instance('Cwis_Defender_Settings');
        $cwis_defender_status = $cwis_defender_settings->get('smartwaf');
        $cwis_gdpr_status = get_option('CWIS_GDPR_ENABLED');
        $cwis_updater_status = get_option('CWIS_CORE_UPDATES') || get_option('CWIS_PLUGIN_UPDATES') ||
                get_option('CWIS_THEME_UPDATES');
        ?>

        <style type="text/css">
            .cobwebsec-dashicon {
                font-size: 18px;
                vertical-align: text-top;
            }
            .cobwebsec-float-right {
                float: right;
            }
            .cobwebsec-text-gray {
                color: #72777c;
            }
            .cobwebsec-text-gray:hover {
                color: #62676c;
            }
            .cobwebsec-widget-header, .cobwebsec-widget-footer {
                margin: 1em -1em;
                padding: 1em;
            }                
            .cobwebsec-widget-header {
                border-bottom: 1px solid #eee;
                padding-top: 0;
            }
            .cobwebsec-widget-header ul {
                margin: 0 0 -6px;
            }
            .cobwebsec-widget-rss ul li {
                padding: 4px 0;
                margin: 0;
            }
            .cobwebsec-widget-rss ul li p {
                margin: 4px 0;
            }
            .cobwebsec-widget-footer {
                color: #ddd;
                border-top: 1px solid #eee;
                padding-bottom: 0;
            }
        </style>

        <div class="cobwebsec-widget-header">
            <ul>
                <li>
                    <a href="<?php echo esc_url($cwis_scanner_url); ?>#/defender/settings">
                        <span class="cobwebsec-dashicon dashicons dashicons-shield"></span>
                        <strong><?php echo __('Smart Protection', CWIS_PLUGIN_SLUG); ?></strong>
                        <span class="cobwebsec-float-right cobwebsec-text-gray"><?php
                            echo __($cwis_defender_status ? 'Enabled' : 'Disabled', CWIS_PLUGIN_SLUG);
                            ?></span>
                    </a>
                </li>
                <li>
                    <a href="<?php echo esc_url($cwis_gdpr_url); ?>">
                        <span class="cobwebsec-dashicon dashicons dashicons-id-alt"></span>
                        <strong><?php echo __('GDPR Regulation Tools', CWIS_PLUGIN_SLUG); ?></strong>
                        <span class="cobwebsec-float-right cobwebsec-text-gray"><?php
                            echo __($cwis_gdpr_status ? 'Enabled' : 'Disabled', CWIS_PLUGIN_SLUG);
                            ?></span>
                    </a>
                </li>
                <li>
                    <a href="<?php echo esc_url($cwis_updater_url); ?>">
                        <span class="cobwebsec-dashicon dashicons dashicons-admin-settings"></span>
                        <strong><?php echo __('WebDefender Updater', CWIS_PLUGIN_SLUG); ?></strong>
                        <span class="cobwebsec-float-right cobwebsec-text-gray"><?php
                            echo __($cwis_updater_status ? 'Active' : 'Inactive', CWIS_PLUGIN_SLUG);
                            ?></span>
                    </a>
                </li>
                <li>
                    <a href="<?php echo esc_url($cwis_scanner_url); ?>">
                        <span class="cobwebsec-dashicon dashicons dashicons-clock"></span>
                        <strong><?php echo __('Last Scanner Activity', CWIS_PLUGIN_SLUG); ?></strong>
                        <span class="cobwebsec-float-right cobwebsec-text-gray"><?php
                            get_option('CWIS_LAST_SCAN') ?
                                            printf(__('%s ago', CWIS_PLUGIN_SLUG),
                                                      human_time_diff(get_option('CWIS_LAST_SCAN'))) :
                                            print(__('N/A', CWIS_PLUGIN_SLUG));
                            ?></span>
                    </a>
                </li>
                <li>
                    <em style="font-size:smaller"><sup>*</sup> Once a week is generally a recommended frequency for a full system scan.
                        This allows your antivirus scanner to run the newest virus definition files against your server system.</em>
                </li>
            </ul>
        </div>

        <div class="cobwebsec-widget-rss">
            <ul>
                <?php if (empty($rss_items)) : ?>
                    <li><?php _e('No items', CWIS_PLUGIN_SLUG); ?></li>
                <?php else : ?>
                    <?php // Loop through each feed item and display each item as a hyperlink.   ?>
                    <?php foreach ($rss_items as $item) : ?>
                        <li>
                            <a href="<?php
                            echo add_query_arg(
                                    array('utm_campaign' => 'feed', 'utm_medium' => 'dashboard_widget'),
                                    esc_url($item['link'])
                            );
                            ?>" target="_blank" title="<?php esc_html($item['date']); ?>"><?php
                                   echo esc_html($item['title']);
                                   ?></a>
                            <?php /* if ($item['description']): ?>
                              <p><?php echo esc_html($item['description']); ?></p>
                              <?php endif; */ ?>
                        </li>
                    <?php endforeach; ?>
                <?php endif; ?>
            </ul>
        </div>

        <p class="cobwebsec-widget-footer">
            <a href="https://wordpress.org/plugins/cwis-antivirus-malware-detected/#developers" target="_blank"
               class="cobwebsec-float-right">
                <span class="cobwebsec-text-gray">Version <?php echo $this->get_version(); ?></span>
                <span class="screen-reader-text">(opens in a new window)</span>
            </a>

            <a href="https://cobweb-security.com/our-product/" target="_blank">Our Products
                <span class="screen-reader-text">(opens in a new window)</span><span aria-hidden="true" class="dashicons dashicons-external"></span>
            </a>
            |

            <a href="https://cobweb-security.com/website-under-attack/" target="_blank">Fix &amp; Protect
                <span class="screen-reader-text">(opens in a new window)</span><span aria-hidden="true" class="dashicons dashicons-external"></span>
            </a>
            |

            <a href="https://cobweb-security.com/blog/" target="_blank">Blog
                <span class="screen-reader-text">(opens in a new window)</span><span aria-hidden="true" class="dashicons dashicons-external"></span>
            </a>
        </p>
        <?php
    }

    /**
     * Retrieves an external feed and parses it. Uses the SimplePie and FeedCache
     * functionality for retrieval and parsing and automatic caching.
     * 
     * @since   4.1.6
     * @access  private
     * 
     * @return  array
     */
    private function get_rss_items()
    {
        if (false === ($rss_items = get_transient('cobwebsec_feed_items'))) {
            $rss_items = array();

            // Get a SimplePie RSS feed object
            include_once( ABSPATH . WPINC . '/feed.php' );
            $rss = fetch_feed($this->feed_uri);
            if (!is_wp_error($rss)) {

                // Figure out how many total items there are, but limit it to 5. 
                $max_items = $rss->get_item_quantity(5);

                // Build an array of all the items, starting with element 0 (first element).
                $items = $rss->get_items(0, $max_items);
                foreach ($items as $item) {
                    $rss_items[] = array(
                        'title' => $item->get_title(),
                        //'description' => wp_trim_words(strip_tags($item->get_description()), 30),
                        'date'  => sprintf(__('Posted %s', CWIS_PLUGIN_SLUG), $item->get_date('j F Y | g:i a')), //$item->get_date('U'),
                        'link'  => $item->get_permalink(),
                    );
                }
                set_transient('cobwebsec_feed_items', $rss_items, 24 * HOUR_IN_SECONDS);
            }
        }

        return $rss_items;
    }

}
