<?php

/**
 * The admin-specific functionality of the plugin.
 * 
 * Defines the plugin name and version, and two examples hooks for how to
 * enqueue the admin-specific stylesheet and JavaScript.
 *
 * @link       https://cobweb-security.com
 * @since      4.2.0
 *
 * @package    Cwis
 * @subpackage Cwis/includes/gdpr
 * @author     Cobweb Security <cwis@cobweb-security.com>
 */
class Cwis_GDPR_TCPDF extends TCPDF
{

    private $plugin_name;
    private $version;
    // TCPDF settings
    private $tcpdf_font_family = 'helvetica';
    private $tcpdf_header_margin = 30;

    public function __construct($plugin_name, $version, $loader = null)
    {
        parent::__construct();
        $this->plugin_name = $plugin_name;
        $this->version = $version;
    }

    /**
     * Custom PDF Page header.
     * 
     * @since   4.2.0
     * @access  public
     */
    public function Header()
    {
        // Set header and font
        $blogname = get_bloginfo('title');
        $titleHtml = '<h3 style="color:#23282d;font-weight:bold;font-size:16px">' . $blogname . '</h3><hr/>';
        $this->SetFont($this->tcpdf_font_family, '', 10);

        $this->writeHTML($titleHtml);
    }

    /**
     * Custom PDF Page footer.
     * 
     * @since   4.2.0
     * @access  public
     */
    public function Footer()
    {
        $blogurl = get_bloginfo('url');

        // Position at 30 mm from bottom
        $this->SetY(-15);

        // Set font
        $this->SetFont($this->tcpdf_font_family, '', 8);
        $footerHtml = '<table border="1" cellpadding="5" cellspacing="0" width="100%">
            <tr>
                <td colspan="2" style="color:#2D2E2E;text-align:center;"><i>' . $blogurl . '</i></td>
            </tr>								
        </table>
        <div style="color:#2D2E2E;text-align:center;line-height:3;">Page ' . $this->getAliasNumPage() . '/' . $this->getAliasNbPages() . '</div>';

        $this->WriteHTML($footerHtml, true, 0, true, 0);
    }

    /**
     * Send the document to a given destination: browser, file or string.
     * In the case of a browser, the PDF viewer may be used or a download may be forced.
     * The method first calls Close() if necessary to terminate the document.
     * 
     * dest - Destination where to send the document (the default value is "S").
     * It can be one of the following:
     *  I: send the file inline to the browser. The PDF viewer is used if available.
     *  D: send to the browser and force a file download with the name given by name.
     *  F: save to a local file with the name given by name (may include a path).
     *  S: return the document as a string.
     * 
     * @since   4.2.0
     * @access  public
     * 
     * @param   string $export_data
     * @param   string $export_filename
     * @param   string $dest
     * 
     * @return  mixed
     */
    public function OutputExportData($export_data = '', $export_filename = '', $dest = 'S')
    {
        ob_start();

        // Set document information
        $this->SetCreator(PDF_CREATOR);

        // Set header and footer fonts
        $this->setHeaderFont(Array(PDF_FONT_NAME_MAIN, '', PDF_FONT_SIZE_MAIN));
        $this->setFooterFont(Array(PDF_FONT_NAME_DATA, '', PDF_FONT_SIZE_DATA));

        // Set default monospaced font
        $this->SetDefaultMonospacedFont(PDF_FONT_MONOSPACED);

        // Set margins
        $_marginTop = $this->tcpdf_header_margin * 0.352777778;  // convert "pixel" value into "mm" (1px = 0.352777778mm for 72dpi)
        $this->SetMargins(PDF_MARGIN_LEFT, $_marginTop + 8, PDF_MARGIN_RIGHT);
        $this->SetHeaderMargin(PDF_MARGIN_HEADER);
        $this->SetFooterMargin(PDF_MARGIN_FOOTER);

        // Set auto page breaks
        $this->SetAutoPageBreak(TRUE, 22);

        // Set image scale factor
        $this->setImageScale(PDF_IMAGE_SCALE_RATIO);

        // ---------------------------------------------------------
        // Add a page
        $this->AddPage('L', 'A4');
        $this->SetFont($this->tcpdf_font_family, '', 10);
        $this->SetTextColor(40, 40, 40);

        // Write HTML data
        $this->writeHTML($export_data, true, false, false, false, '');

        ob_end_clean();
        return $this->Output($export_filename, $dest, true);
    }

}

// Extend the TCPDF class to create custom Header and Footer
class Cwis_GDPR_MyPDF extends TCPDF
{

    //PDF Page header
    public function Header()
    {
        $blogname = get_bloginfo('title');
        $titleHtml = '<h3 style="color:#333333;font-weight:bold;font-size:16px;">' . $blogname . '</h3><hr/>';
        $this->SetFont($this->tcpdf_font_family, '', 10);
        $this->writeHTML($titleHtml);
    }

    // PDF Page footer
    public function Footer()
    {
        $blogurl = get_bloginfo('url');
        // Position at 30 mm from bottom
        $this->SetY(-15);

        // Set font
        $this->SetFont($this->tcpdf_font_family, '', 8);

        $footerHtml = '<table border="1" cellpadding="5" cellspacing="0" width="100%">
								<tr>
									<td colspan="2" style="color:#2D2E2E;text-align:center;"><i>' . $blogurl . '</i></td>
								</tr>								
							</table><div style="color:#2D2E2E;text-align:center;line-height:3;">Page ' . $this->getAliasNumPage() . '/' . $this->getAliasNbPages() . '</div>';
        $this->WriteHTML($footerHtml, true, 0, true, 0);
    }

}
